/*
 * debtags - Implement package tags support for Debian
 *
 * Copyright (C) 2003--2006  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#define APPNAME PACKAGE_NAME
#else
#warning No config.h found: using fallback values
#define APPNAME __FILE__
#define PACKAGE_VERSION "unknown"
#endif

#include <tagcoll/input/stdio.h>
#include <tagcoll/stream/filters.h>
#include <tagcoll/expression.h>
#include <tagcoll/TextFormat.h>
#include <tagcoll/SmartHierarchy.h>
#include <tagcoll/coll/simple.h>
#include <tagcoll/utils/set.h>
//#include <tagcoll/experiments.h>

#include "Environment.h"
#include "DebtagsOptions.h"
#include "Printer.h"
#include "nullstream.h"
#include "Ept.h"
#include "SmartSearcher.h"

#include <ept/cache/debtags/update.h>
#include <ept/cache/debtags/vocabularymerger.h>
//#include <apt-pkg/configuration.h>

#include <wibble/sys/fs.h>

#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/types.h>	// umask
#include <sys/stat.h>	// umask


namespace std {

template<typename TAG, typename _Traits>
basic_ostream<char, _Traits>& operator<<(basic_ostream<char, _Traits>& out, const std::set<TAG>& tags)
{
	for (typename std::set<TAG>::const_iterator i = tags.begin();
			i != tags.end(); i++)
		if (i == tags.begin())
			out << i->fullname();
		else
			out << ", " << i->fullname();
	return out;
}

template<typename TAG, typename _Traits>
basic_ostream<char, _Traits>& operator<<(basic_ostream<char, _Traits>& out, const wibble::Singleton<TAG>& tags)
{
	out << *tags.begin();
	return out;
}

template<typename TAG, typename _Traits>
basic_ostream<char, _Traits>& operator<<(basic_ostream<char, _Traits>& out, const wibble::Empty<TAG>&)
{
	return out;
}

}

using namespace std;

template<typename OUT>
class ExpressionFilter : public wibble::mixin::OutputIterator< ExpressionFilter<OUT> >
{
	tagcoll::Expression expr;
	bool invert;
	OUT out;
public:
	ExpressionFilter(const tagcoll::Expression& expr, bool invert, const OUT& out)
		: expr(expr), invert(invert), out(out) {}

	template<typename ITEMS, typename TAGS>
	ExpressionFilter<OUT>& operator=(const std::pair<ITEMS, TAGS>& data)
	{
		std::set<std::string> tags;
		for (typename TAGS::const_iterator i = data.second.begin();
				i != data.second.end(); ++i)
			tags.insert(i->fullname());
		bool matched = expr(tags);
		if (invert)
			matched = !matched;
		if (matched)
		{
			*out = data;
			++out;
		}
		return *this;
	}
};
template<typename OUT>
ExpressionFilter<OUT> expressionFilter(const tagcoll::Expression& expr, bool invert, const OUT& out)
{
	return ExpressionFilter<OUT>(expr, invert, out);
}


auto_ptr<Ept> debtagsInit(bool write = false)
{
	return auto_ptr<Ept>(new Ept(write));
}

template<typename OUT>
void readCollection(const string& file, const OUT& out)
{
	if (file == "-")
	{
		input::Stdio input(stdin, "<stdin>");
		textformat::parse(input, out);
	}
	else
	{
		input::Stdio input(file);
		textformat::parse(input, out);
	}
}

class SubstringTagMatcher
{
protected:
	vector<string> patterns;

public:
	void add(const std::string& pattern) { patterns.push_back(pattern); }
	
	bool operator()(const Facet& item)
	{
		for (vector<string>::const_iterator i = patterns.begin();
				i != patterns.end(); i++)
		{
			if (strcasestr(item.name().c_str(), i->c_str()))
				return true;
			if (strcasestr(item.shortDescription().c_str(), i->c_str()))
				return true;
			if (strcasestr(item.longDescription().c_str(), i->c_str()))
				return true;
		}
		return false;
	}
	bool operator()(const Tag& item)
	{
		//cerr << "Testing: " << item.fullname() << endl;
		for (vector<string>::const_iterator i = patterns.begin();
				i != patterns.end(); i++)
		{
			if (strcasestr(item.fullname().c_str(), i->c_str()))
				return true;
			if (strcasestr(item.shortDescription().c_str(), i->c_str()))
				return true;
			if (strcasestr(item.longDescription().c_str(), i->c_str()))
				return true;
		}
		//cerr << "  not passed." << endl;
		return false;
	}
};

template<class OUT>
class FacetOnly : public wibble::mixin::OutputIterator< FacetOnly<OUT> >
{
	OUT out;
public:
	FacetOnly(const OUT& out) : out(out) {}

	template<typename ITEMS, typename TAGS>
	FacetOnly<OUT>& operator=(const std::pair<ITEMS, TAGS>& data)
	{
		std::set<std::string> items;
		std::set<std::string> tags;
		for (typename ITEMS::const_iterator i = data.first.begin();
				i != data.first.end(); ++i)
			items.insert(i->name());
		for (typename TAGS::const_iterator i = data.second.begin();
				i != data.second.end(); ++i)
			tags.insert(i->facet().name());
		*out = make_pair(items, tags);
		++out;
		return *this;
	}
};
template<typename OUT>
FacetOnly<OUT> facetOnly(const OUT& out)
{
	return FacetOnly<OUT>(out);
}

// Needs Package to see if it's installed
template<typename OUT>
class TODOFilter : public wibble::mixin::OutputIterator< TODOFilter<OUT> >
{
	Expression nyt;
	OUT out;

public:
	TODOFilter(const OUT& out) : nyt("special::not-yet-tagged"), out(out) {}

	template<typename ITEMS, typename TAGS>
	TODOFilter<OUT>& operator=(const std::pair<ITEMS, TAGS>& data)
	{
		for (typename ITEMS::const_iterator i = data.first.begin();
				i != data.first.end(); ++i)
		{
			if (!(i->isInstalled() || i->isUpgradable() || i->isBroken()))
				continue;
			std::set<string> tags;
			for (typename TAGS::const_iterator j = data.second.begin();
					j != data.second.end(); ++j)
			{
				tags.insert(j->fullname());
			}
			if (nyt(tags))
			{
				*out = make_pair(wibble::singleton(*i), data.second);
				++out;
			}
		}
		return *this;
	}
};

template<typename OUT>
TODOFilter<OUT> todoFilter(const OUT& out)
{
	return TODOFilter<OUT>(out);
}

struct Stats
{
	int seen;
	int onlynyt;
	int nyt;
	int notags;
	int tagged;
	int complete;

	Stats() : seen(0), onlynyt(0), nyt(0), notags(0), tagged(0), complete(0) {}
};

// Might need package in the future to count stats about installed packages
// TODO: count stats about installed packages
class StatsCollector : public wibble::mixin::OutputIterator< StatsCollector >
{
protected:
	Stats& stats;
	Expression has_nyt;
	Expression is_ct;

public:
	StatsCollector(Stats& stats) : stats(stats),
			has_nyt("special::not-yet-tagged*"),
			is_ct("special::completely-tagged") {}

	template<typename ITEMS, typename TAGS>
	StatsCollector& operator=(const std::pair<ITEMS, TAGS>& data)
	{
		++stats.seen;
		if (data.second.empty())
			++stats.notags;
		else {
			std::set<string> tags;
			for (typename TAGS::const_iterator i = data.second.begin();
					i != data.second.end(); ++i)
				tags.insert(i->fullname());

			if (has_nyt(tags)) {
				++stats.nyt;
				++stats.onlynyt;
				for (set<string>::const_iterator i = tags.begin(); i != tags.end(); ++i)
					if (i->substr(0, 23) != "special::not-yet-tagged")
					{
						--stats.onlynyt;
						break;
					}
			} else {
				++stats.tagged;
				if (is_ct(tags))
					++stats.complete;
			}
		}
		return *this;
	}
};

struct VocabularyCheck
{
	int missing_count;
	map<string, int> missing;
	unsigned int mtag_width;
	int max_missing;

	VocabularyCheck() : missing_count(0), mtag_width(0), max_missing(0) {}

	void report(ostream& out)
	{
		if (missing_count > 0)
		{
			out << missing.size() << " tags were found in packages but not in the vocabulary" << endl;
			out << "This happened " << missing_count << " times" << endl;
			out << "The tags found in the collection but not in the vocabulary are:" << endl;

			int mtag_maxdigits = 0;
			for (int i = max_missing; i > 0; i = i / 10)
				mtag_maxdigits++;

			for (map<string, int>::const_iterator i = missing.begin();
					i != missing.end(); i++)
			{
				char buf[1000];
				snprintf(buf, 1000, "\t%s %*s (in %*d package%s)",
					i->first.c_str(), 
					mtag_width - i->first.size(),
					"", 
					mtag_maxdigits,
					i->second,
					i->second > 1 ? "s" : "");
				out << buf << endl;
			}
		}
	}
};

// Check a collection agaisnt a tag vocabulary
class VocabularyChecker : public wibble::mixin::OutputIterator< VocabularyChecker >
{
protected:
	const Vocabulary& voc;
	VocabularyCheck& res;

public:
	VocabularyChecker(const Vocabulary& voc, VocabularyCheck& res) : voc(voc), res(res) {}

	template<typename ITEMS, typename TAGS>
	VocabularyChecker& operator=(const std::pair<ITEMS, TAGS>& data)
	{
		for (typename TAGS::const_iterator i = data.second.begin();
				i != data.second.end(); ++i)
			if (!voc.hasTag(*i))
			{
				const string& tag = *i;
				res.missing_count++;
				res.missing[tag]++;
				if (tag.size() > res.mtag_width)
					res.mtag_width = tag.size();
				if (res.missing[tag] > res.max_missing)
					res.max_missing = res.missing[tag];
			}
		return *this;
	}
};

template<typename VOC, typename OUT>
class VocabularyFilter : public wibble::mixin::OutputIterator< VocabularyFilter<VOC, OUT> >
{
protected:
	const VOC& voc;
	OUT out;

public:
	VocabularyFilter(const VOC& voc, const OUT& out) : voc(voc), out(out) {}

	template<typename ITEMS, typename TAGS>
	VocabularyFilter& operator=(const std::pair<ITEMS, TAGS>& data)
	{
		std::set< typename TAGS::value_type > tags;
		for (typename TAGS::const_iterator i = data.second.begin();
				i != data.second.end(); ++i)
			if (voc.hasTag(*i))
				tags.insert(*i);
		*out = make_pair(data.first, tags);
		++out;
		return *this;
	}
};
template<typename VOC, typename OUT>
VocabularyFilter<VOC, OUT> vocabularyFilter(const VOC& voc, const OUT& out)
{
	return VocabularyFilter<VOC, OUT>(voc, out);
}

static void printShortVocabularyItem(const Facet& facet)
{
	cout << facet.name() << " (facet) - " << facet.shortDescription() << endl;
}

static void printShortVocabularyItem(const Tag& tag)
{
	cout << tag.fullname() << " - " << tag.shortDescription() << endl;
}

static void printVocabularyItem(const Facet& tag)
{
	string ld = tag.longDescription();
	cout << "Facet: " << tag.name() << endl;
	cout << "Description: " << tag.shortDescription() << endl;
	cout << " " << ld << endl;
	if (ld[ld.size() - 1] != '\n')
		cout << endl;
}
static void printVocabularyItem(const Tag& tag)
{
	string ld = tag.longDescription();
	cout << "Tag: " << tag.fullname() << endl;
	cout << "Description: " << tag.shortDescription() << endl;
	cout << " " << ld << endl;
	if (ld[ld.size() - 1] != '\n')
		cout << endl;
}

template<typename OUT>
class PackageToMaint : public wibble::mixin::OutputIterator< PackageToMaint<OUT> >
{
	OUT out;
public:
	PackageToMaint(const OUT& out) : out(out) {}

	template<typename ITEMS, typename TAGS>
	PackageToMaint& operator=(const std::pair<ITEMS, TAGS>& data)
	{
		std::set<string> items;
		std::set<string> tags;
		for (typename ITEMS::const_iterator i = data.first.begin();
				i != data.first.end(); ++i)
			if (i->valid())
			{
				Version v = i->anyVersion();
				if (v.valid())
					items.insert(v.maintainer());
			}
		for (typename TAGS::const_iterator i = data.second.begin();
				i != data.second.end(); ++i)
			if (i->valid())
				tags.insert(i->fullname());
		if (!items.empty())
		{
			*out = make_pair(items, tags);
			++out;
		}
		return *this;
	}
};
template<typename OUT>
PackageToMaint<OUT> packageToMaint(const OUT& out)
{
	return PackageToMaint<OUT>(out);
}


struct PackageScore
{
	// tag -> (installed, total)
	map<Tag, pair<int, int> > tagScore;
	set<Package> uninstalled;

	void output(TagMap& tagmap, ostream& out)
	{
		map<int, Package> sorter;

		for (set<Package>::const_iterator i = uninstalled.begin();
				i != uninstalled.end(); ++i)
		{
			// Compute package score
			set<Tag> tags = tagmap.getTagsOfItem(*i);
			int score = 0;
			int ntags = 0;
			for (set<Tag>::const_iterator t = tags.begin();
					t != tags.end(); ++t)
			{
				map<Tag, pair<int, int> >::const_iterator j = tagScore.find(*t);
				if (j != tagScore.end())
				{
					score += 100 * j->second.first / j->second.second;
					++ntags;
				}
			}
			sorter.insert(make_pair(score / ntags, *i));
		}

		for (map<int, Package>::const_iterator i = sorter.begin();
				i != sorter.end(); ++i)
			out << i->first << " " << i->second.name() << " - " <<
				i->second.shortDescription(string("(short description not available)")) << endl;
	}
};

// Needs package for TagsScorer
class PackageScorer : public wibble::mixin::OutputIterator< PackageScorer >
{
protected:
	tagcoll::Expression avoidPkg;
	tagcoll::Expression avoidTag;
	PackageScore& res;

public:
	PackageScorer(PackageScore& res)
		: avoidPkg("role::sw:shlib || role::content:data"), avoidTag("special::*"), res(res) {}

	template<typename ITEMS, typename TAGS>
	PackageScorer& operator=(const std::pair<ITEMS, TAGS>& data)
	{
		// Skip items that need skipping
		set<string> stags;
		for (typename TAGS::const_iterator i = data.second.begin();
				i != data.second.end(); ++i)
			stags.insert(i->fullname());
		if (avoidPkg(stags))
			return *this;

		for (typename ITEMS::const_iterator i = data.first.begin();
				i != data.first.end(); ++i)
		{
			bool installed = i->isInstalled() || i->isUpgradable() || i->isBroken();

			if (!installed)
				// Take note of uninstalled packages
				res.uninstalled.insert(*i);

			// Score tags of installed packages
			for (typename TAGS::const_iterator t = data.second.begin();
					t != data.second.end(); ++t)
			{
				std::set<string> it;
				it.insert(t->fullname());
				if (!avoidTag(it))
				{
					cerr << "Count " << t->fullname() << endl;
					map<Tag, pair<int, int> >::iterator j = res.tagScore.find(*t);
					if (j == res.tagScore.end())
						res.tagScore.insert(make_pair(*t, make_pair(installed ? 1 : 0, 1)));
					else
					{
						if (installed)
							++j->second.first;
						++j->second.second;
					}
				}
			}
		}
		return *this;
	}
};

#if 0
template<class ITEM>
class FacetcollPrinter : public Tagcoll::Consumer<ITEM, Facet>
{
protected:
	virtual void consumeItemUntagged(const ITEM& item)
	{
		printf("%.*s\n", PFSTR(item));
	}
	virtual void consumeItem(const ITEM& item, const OpSet<Facet>& tags)
	{
		printf("%.*s: ", PFSTR(item));
		for (OpSet<Facet>::const_iterator i = tags.begin();
				i != tags.end(); i++)
			if (i == tags.begin())
				printf("%.*s", PFSTR(i->name()));
			else
				printf(", %.*s", PFSTR(i->name()));
		printf("\n");

	}

public:
	virtual ~FacetcollPrinter() {}
};

template<>
void FacetcollPrinter<entity::Package>::consumeItemUntagged(const entity::Package& item)
{
	printf("%.*s\n", PFSTR(item.name()));
}
template<>
void FacetcollPrinter<entity::Package>::consumeItem(const entity::Package& item, const OpSet<Facet>& tags)
{
	printf("%.*s: ", PFSTR(item.name()));
	for (OpSet<Facet>::const_iterator i = tags.begin();
			i != tags.end(); i++)
		if (i == tags.begin())
			printf("%.*s", PFSTR(i->name()));
		else
			printf(", %.*s", PFSTR(i->name()));
	printf("\n");

}

template< typename PKG, typename TAG >
class TagcollConsumerAdaptor : public utils::ConsumerImpl< PKG, TagcollConsumerAdaptor<PKG, TAG> >
{
protected:
	Consumer<PKG, TAG>& m_out;

public:
	TagcollConsumerAdaptor( Consumer<PKG, TAG>& out) : m_out( out ) {}
	virtual void consume( const PKG& a ) {
		if (a != PKG())
			m_out.consume(a, a.tags());
	}
};

template< typename ITEM >
class RangeCounter : public utils::ConsumerImpl< ITEM, RangeCounter<ITEM> >
{
protected:
	int* count;

public:
	RangeCounter(int* count) : count(count) {}
	virtual void consume(const ITEM& a) {
		++*count;
	}
	
};

class Searcher
{
protected:
	component::PackageTags& debtags;
	Printer<Package, Tag>* printer;

public:
	Searcher(component::PackageTags& debtags, Printer<Package, Tag>* printer) :
		debtags(debtags), printer(printer) {}

	void output()
	{
		debtags.outputPatched(*printer);
		printer->flush();
	}

	int output(const string& expression, bool invertMatch)
	{
		// Build the grep filter chain
		FilterItemsByExpression<Package, Tag> filter(*printer, expression);
		if (invertMatch)
			filter.setMatchType(FilterItemsByExpression<Package, Tag>::INVERTED);
		debtags.outputPatched(filter);
		printer->flush();
		return filter.countMatched();
	}

	int output(predicate::Predicate<Package> p)
	{
		aptFront::cache::component::Packages& packages = Global::get().packages();
		utils::Range<Package> r = filteredRange(
				utils::Range<Package>(range(packages.packagesBegin(), packages.packagesEnd())), p);
		r.output(TagcollConsumerAdaptor<Package, Tag>(*printer));
		printer->flush();
		int count = 0;
		RangeCounter<Package> counter(&count);
		r.output(counter);
		return count;
	}
};
#endif


#if 0
template<class ITEM>
class ExprCollector : public Tagcoll::Consumer<ITEM, Tag>, public OpSet<entity::Package>
{
protected:
	Expression expr;

	virtual void consumeItemUntagged(const entity::Package& item) throw ()
	{
		if (expr(OpSet<Tag>()))
			(*this) += item;
	}
	virtual void consumeItem(const entity::Package& item, const OpSet<Tag>& tags) throw ()
	{
		if (expr(tags))
			(*this) += item;
	}

public:
	ExprCollector(const std::string& expression)
		: expr(expression) {}
	virtual ~ExprCollector() {}
};
#endif

#if 0
template<class ITEM>
OpSet<ITEM> getItems(HierarchyNode<ITEM, Facet>* node)
{
	OpSet<ITEM> items = node->getItems();
	
	for (typename HierarchyNode<ITEM, Facet>::iterator i = node->begin();
			i != node->end(); i++)
		items += getItems(*i);

	return items;
}

// Collect items that are making a potential implication between 'existing' and
// 'missing' fail.  Basically, collect all items that have facet 'existing' but
// not facet 'missing'
template<class ITEM>
class MICollector : public Tagcoll::Consumer<ITEM, Facet>
{
protected:
	const Facet& existing;
	const Facet& missing;
	OpSet<ITEM> found;
public:
	MICollector(const Facet& existing, const Facet& missing) : existing(existing), missing(missing) {}
	virtual ~MICollector() throw() {}

	virtual void consume(const ITEM& item) throw () {}
	virtual void consume(const OpSet<ITEM>& items) throw () {}

	virtual void consume(const ITEM& item, const OpSet<Facet>& tags) throw ()
	{
		if (tags.contains(existing) && !tags.contains(missing))
			found += item;
	}

	virtual void consume(const OpSet<ITEM>& items, const OpSet<Facet>& tags) throw ()
	{
		if (tags.contains(existing) && !tags.contains(missing))
			found += items;
	}

	const OpSet<ITEM>& getFound() const throw () { return found; }
};


template<class ITEM>
OpSet<entity::Package> getMissedImplications(const Collection<ITEM, Facet>& coll,
									const Facet& existing, const Facet& missing)
{
	MICollector<ITEM> mc(existing, missing);
	coll.output(mc);
	return mc.getFound();
}

// Needs Package to get the short description
// TODO: just instantiate the package when it actually needs the short
// description
class ReportMaker
{
protected:
	component::PackageTags& debtags;
	OpSet<entity::Package> mentionedPackages;
	unsigned int itemsPerGroup;

	void listSuspects(const OpSet<entity::Package>& pkgs)
	{
		OpSet<entity::Package> toPrint = pkgs - mentionedPackages;
		if (itemsPerGroup > 0 && itemsPerGroup < toPrint.size())
		{
			float step = (float)toPrint.size() / (float)(itemsPerGroup + 1);
			int count = 0;
			float next = step;
			for (OpSet<entity::Package>::const_iterator i = toPrint.begin(); i != toPrint.end(); i++, count++)
				if (count == (int)next)
				{
					printf(" * `%.*s`_: %.*s\n", PFSTR(i->name()), PFSTR(i->shortDescription(string("(short description not available)"))));
					mentionedPackages += *i;
					next += step;
					if (next > toPrint.size())
						break;
				}
		} else {
			for (OpSet<entity::Package>::const_iterator i = toPrint.begin(); i != toPrint.end(); i++)
				printf(" * `%.*s`_: %.*s\n", PFSTR(i->name()), PFSTR(i->shortDescription(string("(short description not available)"))));
			mentionedPackages += toPrint;
		}
		fputs("\n", stdout);
	}

	void out(const char* text) { fputs(text, stdout); }
	void outTitle()
	{
		out(
			"Debtags TODO Report\n"
			"===================\n"
			"This is an automatically-generated report of Debian packages for which tagging seems\n"
			"to still be incomplete.  It does not aim at being a complete report, but rather it\n"
			"aims at collecting incompleteness that can be detected automatically.\n"
			"\n"
			"To make this list shorter, packages needing work are listed only once.  It might thus\n"
			"happen than when a problem is solved in a package, a further report will still list it,\n"
			"but in a different section.\n"
			"\n"
			"For more informations, please visit the Debtags website at http://debtags.alioth.debian.org\n"
			"\n"
		);
	}
	void outSpecials()
	{
		out(
			"Special-case packages\n"
			"---------------------\n"
			"Previous section: `Debtags TODO Report`_; next section: `Failed implications`_.\n"
			"\n"
			"This section lists packages whose tag set is 'strange' compared to the other packages.\n"
			"\n"
			"The list corresponds to those packages reported by ``debtags facetcoll | tagcoll findspecials``\n"
			"as being in groups with less than 100 items:\n"
			"\n"
		);

		CardinalityStore<entity::Package, Facet> coll;
		TagToFacet<entity::Package> tagStripper(coll);
		debtags.outputPatched(tagStripper);
		Facet f;
		SmartHierarchyNode<entity::Package, Facet> node(f, coll, 0);

		OpSet<Facet> seen;
		OpSet<entity::Package> specials;
		for (HierarchyNode<entity::Package, Facet>::iterator i = node.begin();
				i != node.end(); i++)
		{
			OpSet<entity::Package> items = getItems(*i);

			// Find the items in this branch that are not present in
			// any of the previous ones
			OpSet<entity::Package> newItems;
			if (!seen.empty())
			{
				for (OpSet<entity::Package>::const_iterator j = items.begin();
						j != items.end() && newItems.size() <= 101; j++)
				{
					OpSet<Tag> tags = debtags.tagdb().getTags(*j);
					bool isSeen = false;
					for (OpSet<Tag>::const_iterator i = tags.begin(); i != tags.end(); i++)
						if (seen.contains(i->facet()))
						{
							isSeen = true;
							break;
						}
					if (!isSeen && j->valid())
						newItems += *j;
				}

				if (newItems.size() <= 100)
					specials += newItems;
			}

			seen += (*i)->tag();
		}
		listSuspects(specials);
	}

	void outByExpression(const char* expr, const char* descr)
	{
		out(".. _"); out(expr); out(":\n\n");
		out(descr);
		ExprCollector<entity::Package> miss(expr);
		debtags.outputPatched(miss);
		listSuspects(miss);
	}

	void outImplications()
	{
		out(
			"Failed implications\n"
			"-------------------\n"
			"Previous section: `Special-case packages`_; next section: `Missing general tags`_.\n"
			"\n"
			"This section lists packages that have some tags A which usually go together with some tags B,\n"
			"but they don't actually any tag B.\n"
			"\n"
			"\"Why don't you fix this automatically?\", you could say.  Well, when that's about single tags,\n"
			"that can be automatically done already.  However, in this case we are looking at facets.\n"
			"For example, we are looking at packages which have uitoolkit::* tags but have no interface::*\n"
			"tags.  You can see that automatically adding the interface::* tags is an impossible task.\n"
			"\n"
		);
		outByExpression(
				"uitoolkit::* && ! interface::*",
				"Packages which have 'uitoolkit' but not 'interface' (next: `uitoolkit::* && ! implemented-in::*`_):\n\n"
		);
		outByExpression(
				"uitoolkit::* && ! implemented-in::*",
				"Packages which have 'uitoolkit' but not 'implemented-in' (previous: `uitoolkit::* && ! interface::*`_;\n"
				"next: `Missing general tags`_):\n\n"
		);
		// TODO: find more such couples
	}

	void outMissingGenerals()
	{
		out(
			"Missing general tags\n"
			"--------------------\n"
			"Previous section: `Failed Implications`; next section: `Conclusion`_.\n"
			"\n"
			"This section lists packages that are missing tags from a group that is general enough to apply\n"
			"to all packages.\n"
			"\n"
		);
		outByExpression(
				"! role::*",
				"Packages which have to ``role::`` tags (next: `Conclusion`_):\n\n"
		);
		// TODO: find more such global facets
	}
	void outConclusion()
	{
		out(
			"Conclusion\n"
			"----------\n"
			"The report ends here.  If you have ideas for more automated check, please write them\n"
			"to debtags-devel@lists.alioth.debian.org.\n"
			"\n"
			"If you are reading an HTML version of this report, don't hesitate to click on the names\n"
			"of the packages and fix the problem yourself in Erich's packagebrowser.\n"
			"\n"
			"This report has been generated by ``debtags todoreport``.\n"
		);
	}

public:
	ReportMaker(unsigned int itemsPerGroup = 0) :
		debtags(Global::get().debtags()),
		itemsPerGroup(itemsPerGroup) {}

	void printReport()
	{
		outTitle();
		outSpecials();
		outImplications();
		outMissingGenerals();

		// TODO: all packages with tags that make it assume they
		//       contain software (such as: uitoolkit), but with no
		//       implemented-in tags

		outConclusion();

		out("\n");

		// Print the links to visited packages
		for (OpSet<entity::Package>::const_iterator j = mentionedPackages.begin(); j != mentionedPackages.end(); j++)
			printf(".. _%.*s: http://debian.vitavonni.de/packagebrowser/edit.cgi?package=%.*s\n", PFSTR(j->name()), PFSTR(j->name()));
	}
};
#endif

struct PackageToString {
	string operator()(const Package& pkg) { return pkg.name(); }
} serPackage;
struct TagToString {
	string operator()(const Tag& tag) { return tag.fullname(); }
} serTag;

int main(int argc, const char* argv[])
{
	wibble::commandline::DebtagsOptions opts;

	try {
		// Install the handler for unexpected exceptions
		wibble::exception::InstallUnexpected installUnexpected;

		if (opts.parse(argc, argv))
			return 0;

		if (opts.out_verbose->boolValue())
			::Environment::get().verbose(true);

		if (opts.out_debug->boolValue())
			::Environment::get().debug(true);

		if (opts.foundCommand() == opts.selfcheck)
		{
			auto_ptr<Ept> ept = debtagsInit();
			Vocabulary& voc = ept->vocabulary();
			Index& pkgs = ept->packages();

			// ensure that all facets are readable
			std::set<Facet> facets = voc.facets();
			for (std::set<Facet>::const_iterator i = facets.begin(); i != facets.end(); i++)
			{
				cout << "Checking facet " << i->name(string("<invalid name>")) << "..." << endl;
				i->name(string("foo"));
				i->shortDescription(string("foo"));
				i->longDescription(string("foo"));
				i->tags();
			}

			// ensure that all tags are readable
			std::set<Tag> tags = voc.tags();
			for (std::set<Tag>::const_iterator i = tags.begin(); i != tags.end(); i++)
			{
				cout << "Checking tag " << i->fullname(string("<invalid name>")) << "..." << endl;
				i->name(string("foo"));
				i->fullname(string("foo"));
				i->shortDescription(string("foo"));
				i->longDescription(string("foo"));
			}

			// ensure that all packages that declare they are readable, are readable
			for (Index::iterator i = pkgs.begin(); i != pkgs.end(); ++i)
			{
				cout << "Checking package " << i->name(string("(invalid)")) << "..." << endl;
				if (i->valid())
				{
					i->shortDescription(string("foo"));
					i->longDescription(string("foo"));
					//i->tags(std::set<Tag>());
				}
			}
			
			for (std::set<Facet>::const_iterator i = facets.begin();
					i != facets.end(); i++)
			{
				if (!voc.hasFacet(i->name()))
					cerr << "Vocabulary is not sure about having facet " << i->name() << endl;

				std::set<Tag> tags = i->tags();
				for (std::set<Tag>::const_iterator j = tags.begin();
						j != tags.end(); j++)
					if (!voc.hasTag(j->fullname()))
						cerr << "Vocabulary is not sure about having tag " << j->fullname() << endl;
			}
			return 0;
		}
		// Output the full package tag database
		else if (opts.foundCommand() == opts.cat)
		{
			auto_ptr<Ept> ept = debtagsInit();
			TagMap& debtags = ept->tagmap();
			int count = 0;

			auto_ptr<CollPrinter> printer;

			if (opts.out_names->boolValue())
				printer.reset(new CollPrinter(CollPrinter::NAME, count));
			else if (opts.out_facets->boolValue())
				printer.reset(new CollPrinter(CollPrinter::FACETS, count));
			else if (opts.out_quiet->boolValue())
				printer.reset(new CollPrinter(CollPrinter::QUIET, count));
			else 
				printer.reset(new CollPrinter(CollPrinter::TAGS, count));

			if (opts.hasNext())
			{
				debtags.output(expressionFilter(opts.next(), opts.match_invert->boolValue(), *printer));
			} else
				debtags.output(*printer);

			return count > 0 ? 0 : 1;
		}
		// Output the full package database
		else if (opts.foundCommand() == opts.dumpavail)
		{
			auto_ptr<Ept> ept = debtagsInit();
			int count = 0;

			auto_ptr<PackagePrinter> printer;
			if (opts.out_names->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::NAME, count));
			else if (opts.out_short->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::SHORT, count));
			else if (opts.out_quiet->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::QUIET, count));
			else 
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::FULL, count));

			if (opts.hasNext())
			{
				ept->tagmap().output(expressionFilter(opts.next(), opts.match_invert->boolValue(), *printer));
			} else
				ept->tagmap().output(*printer);

			return count > 0 ? 0 : 1;
		}
		// search [-v] <tag expression>\n"
		// Output the names and description of the packages that match\n"
		// the given tag expression\n"
		else if (opts.foundCommand() == opts.search)
		{
			// TODO: complain if no expression found
			auto_ptr<Ept> ept = debtagsInit();
			int count = 0;

			auto_ptr<PackagePrinter> printer;
			if (opts.out_names->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::NAME, count));
			else if (opts.out_full->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::FULL, count));
			else if (opts.out_quiet->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::QUIET, count));
			else 
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::SHORT, count));

			if (opts.hasNext())
			{
				ept->tagmap().output(expressionFilter(opts.next(), opts.match_invert->boolValue(), *printer));
			} else
				ept->tagmap().output(*printer);

			return count > 0 ? 0 : 1;
		}
		// grep [-v] [-q] <tag expression>
		// Output the lines of the full package tag database that match the
		// given tag expression
		else if (opts.foundCommand() == opts.grep)
		{
			// TODO: complain if no expression found
			auto_ptr<Ept> ept = debtagsInit();
			TagMap& debtags = ept->tagmap();
			int count = 0;

			auto_ptr<CollPrinter> printer;

			if (opts.out_names->boolValue())
				printer.reset(new CollPrinter(CollPrinter::NAME, count));
			else if (opts.out_facets->boolValue())
				printer.reset(new CollPrinter(CollPrinter::FACETS, count));
			else if (opts.out_quiet->boolValue())
				printer.reset(new CollPrinter(CollPrinter::QUIET, count));
			else 
				printer.reset(new CollPrinter(CollPrinter::TAGS, count));

			if (opts.hasNext())
			{
				debtags.output(expressionFilter(opts.next(), opts.match_invert->boolValue(), *printer));
			} else
				debtags.output(*printer);

			return count > 0 ? 0 : 1;
		}
		// tagcat
		// Output the entire tag vocabulary
		else if (opts.foundCommand() == opts.tagcat)
		{
			auto_ptr<Ept> ept = debtagsInit();
			Vocabulary& voc = ept->vocabulary();

			std::set<Facet> facets = voc.facets();
			for (std::set<Facet>::const_iterator i = facets.begin();
					i != facets.end(); i++)
			{
				printVocabularyItem(*i);

				std::set<Tag> tags = i->tags();
				for (std::set<Tag>::const_iterator j = tags.begin();
						j != tags.end(); j++)
					printVocabularyItem(*j);
			}
			return 0;
		}
		// tagshow <tag>
		// Show the vocabulary informations about a tag
		else if (opts.foundCommand() == opts.tagshow)
		{
			auto_ptr<Ept> ept = debtagsInit();

			string tag = opts.next();

			Tag t = ept->vocabulary().tagByName(tag);
			if (!t.valid())
			{
				verbose("Tag `%s' was not found in tag vocabulary\n", tag.c_str());
				return 1;
			}
			else
			{
				printVocabularyItem(t);
				return 0;
			}
		}
		// tagsearch <pattern [pattern [pattern [...]]]>
		// Show a summary of all tags matching the given patterns
		else if (opts.foundCommand() == opts.tagsearch)
		{
			auto_ptr<Ept> ept = debtagsInit();

			SubstringTagMatcher match;

			// Get the patterns to be matched
			bool empty;
			while (opts.hasNext())
			{
				string pattern = opts.next();
				match.add(pattern);
				empty = false;
			}

			if (empty)
			{
				error("No patterns given in commandline\n");
				return 1;
			}

			int matched = 0;

			std::set<Facet> facets = ept->vocabulary().facets();
			for (std::set<Facet>::const_iterator i = facets.begin();
					i != facets.end(); i++)
			{
				if (match(*i))
				{
					matched++;
					printShortVocabularyItem(*i);
				}

				std::set<Tag> tags = i->tags();
				for (std::set<Tag>::const_iterator j = tags.begin();
						j != tags.end(); j++)
					if (match(*j))
					{
						matched++;
						printShortVocabularyItem(*j);
					}
			}

			return matched > 0 ? 0 : 1;
		}
		// show <pkg>
		// Call apt-cache show <pkg>, but add tag informations to the output.\n"
		else if (opts.foundCommand() == opts.show)
		{
			auto_ptr<Ept> ept = debtagsInit();

			while (opts.hasNext())
			{
				string name = opts.next();

				Package pkg = ept->packages().packageByName(name);
				if (pkg.valid())
				{
					cout << pkg.candidateVersion().completeRecord() << endl;
					cout << "Tags: ";
					std::set<Tag> ts = ept->tagmap().getTagsOfItem(pkg);
					for (std::set<Tag>::const_iterator i = ts.begin();
							i != ts.end(); i++)
						if (i == ts.begin())
							cout << i->fullname();
						else
							cout << ", " + i->fullname();
					cout << endl;
					return 0;
				} else {
					verbose("Package %s not found", name.c_str());
					return 1;
				}
			}
		}
		// related <pkg1[,pkg2[,pkg2,[...]]]>
		// Show packages related to the specified ones
		else if (opts.foundCommand() == opts.related)
		{
			using namespace wibble::operators;

			auto_ptr<Ept> ept = debtagsInit();

			int maxdist = 0;
			if (opts.misc_distance->boolValue())
				maxdist = opts.misc_distance->intValue();
			string pkg = opts.next();

			// Split the items on commas
			string splititem;
			std::set<Package> splititems;
			for (string::const_iterator c = pkg.begin(); c != pkg.end(); c++)
				if (*c == ',')
				{
					Package p = ept->packages().packageByName(splititem);
					if (p.valid())
					{
						splititems.insert(p);
					} else {
						error("Item \"%s\" does not exist in the collection\n", splititem.c_str());
						return 1;
					}
					splititem = string();
				} else
					splititem += *c;
			Package p = ept->packages().packageByName(splititem);
			if (p.valid())
			{
				splititems.insert(p);
			} else {
				error("Item \"%s\" does not exist in the collection\n", splititem.c_str());
				return 1;
			}

			// Get the tagset as the intersection of the tagsets of all input items
			std::set<Package>::const_iterator i = splititems.begin();
			std::set<Tag> ts = ept->tagmap().getTagsOfItem(*i);
			for (++i; i != splititems.end(); i++)
				ts &= ept->tagmap().getTagsOfItem(*i);

			if (ts.empty())
			{
				if (splititems.size() > 1)
					fprintf(stderr, "The packages %s are unrelated: cannot find a barycenter to start computing relationships from.\n", pkg.c_str());
				else
					fprintf(stderr, "The package %s has no tags attached.\n", pkg.c_str());
				return 1;
			}

			std::set<Package> related(ept->tagmap().getRelatedItems(ts, maxdist) - splititems);

			int count = 0;

			auto_ptr<PackagePrinter> printer;
			if (opts.out_names->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::NAME, count));
			else if (opts.out_full->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::FULL, count));
			else if (opts.out_quiet->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::QUIET, count));
			else 
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::SHORT, count));

			for (std::set<Package>::const_iterator i = related.begin();
					i != related.end(); ++i)
			{
				**printer = *i;
				++*printer;
			}

			if (count > 50 && maxdist == 0 && isatty(1))
			{
				string tags;
				for (std::set<Tag>::const_iterator i = ts.begin();
						i != ts.end(); i++)
					if (i == ts.begin())
						tags += i->fullname();
					else
						tags += "%2C" + i->fullname();
				feedback("\nIt seems that this set of packages lacks tag information that could help to better distinguish package similarities.\nYou can help providing better tagging: just point your browser to http://debian.vitavonni.de/packagebrowser/?tags=%s\n", tags.c_str());
			}
			return 0;
		}
		// maintainers
		// Create a tagged collection of maintainers and the tags of the
		// packages they maintain
		else if (opts.foundCommand() == opts.maintainers)
		{
			auto_ptr<Ept> ept = debtagsInit();

			// Gather maintainer informations
			coll::Simple<string, string> maints;
			ept->tagmap().output(packageToMaint(inserter(maints)));

			// Print them out
			maints.output(textformat::StdioWriter(stdout));
		}
		// tag
		//   tag [add <package> <tags...>\n"
		//   tag [rm  <package> <tags...>\n"
		//   tag [ls  <package>\n"
		//                View and edit the tags for a package\n");
		else if (opts.foundCommand() == opts.tag)
		{
			std::string cmd = opts.next();

			if (cmd == "add" || cmd == "rm")
			{
				auto_ptr<Ept> ept = debtagsInit(true);

				string name = opts.next();

				Package pkg = ept->packages().packageByName(name);
				if (!pkg.valid())
				{
					error("Package %s not found\n", name.c_str());
					return 1;
				}

				std::set<Tag> tagset;

				while (opts.hasNext())
				{
					string tag = opts.next();
					Tag t = ept->vocabulary().tagByName(tag);
					if (t)
						tagset.insert(t);
					else
						error("Tag '%s' not found: ignored\n", tag.c_str());
				}

				if (!tagset.empty())
				{
					PatchList<Package, Tag> change;
					if (cmd == "add")
						change.addPatch(Patch<Package, Tag>(pkg, tagset, std::set<Tag>()));
					else
						change.addPatch(Patch<Package, Tag>(pkg, std::set<Tag>(), tagset));
					ept->tagmap().applyChange(change);
					ept->tagmap().savePatch();
				} else
					verbose("No tags to add\n");
			}
			else if (cmd == "ls")
			{
				auto_ptr<Ept> ept = debtagsInit();

				string name = opts.next();
				Package pkg = ept->packages().packageByName(name);
				if (pkg.valid())
				{
					std::set<Tag> ts = ept->tagmap().getTagsOfItem(pkg);
					for (std::set<Tag>::const_iterator i = ts.begin();
							i != ts.end(); i++)
						cout << i->fullname() << endl;
					return 0;
				} else {
					verbose("Package %s not found", name.c_str());
					return 1;
				}
			}
			else
				throw wibble::exception::Consistency("parsing the 'tag' subcommand", "command " + cmd + " is not valid working with tags");
		}
		// submit
		// Mail the local updates to the tag database to the central tag
		// repository
		else if (opts.foundCommand() == opts.submit)
		{
			auto_ptr<Ept> ept = debtagsInit();

			if (opts.hasNext())
			{
				input::Stdio in(opts.next());
				PatchList<string, string> patch;
				textformat::parsePatch(in, inserter(patch));
				ept->tagmap().sendPatch(patch);
			}
			else
				ept->tagmap().sendPatch();

		}
		// todo
		// Print a list of the installed packages that are not yet tagged
		else if (opts.foundCommand() == opts.todo)
		{
			auto_ptr<Ept> ept = debtagsInit();
			int count = 0;

			auto_ptr<PackagePrinter> printer;
			if (opts.out_names->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::NAME, count));
			else if (opts.out_full->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::FULL, count));
			else if (opts.out_quiet->boolValue())
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::QUIET, count));
			else 
				printer.reset(new PackagePrinter(ept->tagmap(), PackagePrinter::SHORT, count));

			ept->tagmap().output(todoFilter(*printer));

			return count > 0 ? 0 : 1;
		}
		// stats
		// Print statistics about Debtags
		else if (opts.foundCommand() == opts.stats)
		{
			auto_ptr<Ept> ept = debtagsInit();

			int pkgCount = ept->packages().packageCount();
			printf("Total count of packages: %d\n", pkgCount);

			Stats stats;
			ept->tagmap().outputPatched(StatsCollector(stats));

			printf("Total count of packages (according to APT): %d\n", pkgCount);
			printf("Total count of packages (according to Debtags): %d\n", stats.seen);

			printf("Number of facets: %d\n", ept->vocabulary().facets().size());
			printf("Number of tags: %d\n", ept->vocabulary().tags().size());

#if 0
			// Copied from Debtags class: compute the toplevel facets
			// TODO: use Debtags instead of Environment throughout all Debtags
			CardinalityStore<entity::Package, Facet> coll;
			TagToFacet<entity::Package> tagStripper(coll);
			debtags.outputPatched(tagStripper);
			Facet f;
			SmartHierarchyNode<entity::Package, Facet> node(f, coll, 0);
			printf("Number of automatically computed toplevel facets: %d\n", node.size());
#endif

//			printf("Number of packages with special::completely-tagged tags: %d (%.1f%%)\n",
//					stats.complete, (float)stats.complete*100/stats.seen);
			printf("Number of packages with tags, but no special::not-yet-tagged tags: %d (%.1f%%)\n",
					stats.tagged, (float)stats.tagged*100/stats.seen);
			printf("Number of packages with special::not-yet-tagged tags: %d (%.1f%%)\n",
					stats.nyt, (float)stats.nyt*100/stats.seen);
			printf("Number of packages with only special::not-yet-tagged tags: %d (%.1f%%)\n",
					stats.onlynyt, (float)stats.onlynyt*100/stats.seen);
			printf("Number of packages with no tags: %d (%.1f%%)\n",
					stats.notags, (float)stats.notags*100/stats.seen);

		}
		// check <file>
		// Check that all the tags in the given tagged collection are
		// present in the tag vocabulary.  Checks the main database if no
		// file is specified
		else if (opts.foundCommand() == opts.check)
		{
			auto_ptr<Ept> ept = debtagsInit();

			if (!opts.hasNext())
				throw wibble::exception::BadOption("you should specify the file with the collection to check");

			string file = opts.next();

			VocabularyCheck results;
			ept->readCollectionRaw(file, VocabularyChecker(ept->vocabulary(), results));

			if (results.missing_count > 0)
			{
				results.report(cout);
				return 1;
			}
			else
				return 0;
		}
		// score
		// Score uninstalled packages according to how often their tags
		// appear in the packages that are installed already
		else if (opts.foundCommand() == opts.score)
		{
			auto_ptr<Ept> ept = debtagsInit();

			// Compute package scores
			PackageScore score;
			ept->tagmap().outputPatched(PackageScorer(score));

			// Print the results
			score.output(ept->tagmap(), cout);
		}
		// mkpatch [filename]
		// Create a tag patch between the current tag database and the tag
		// collection [filename]
		else if (opts.foundCommand() == opts.diff)
		{
			auto_ptr<Ept> ept = debtagsInit();

			string file = opts.next();

			coll::Simple<Package, Tag> coll;
			ept->readCollection(file, inserter(coll));

			PatchList<Package, Tag> newpatches;
			newpatches.addPatch(ept->tagmap(), coll);

			textformat::outputPatch(serPackage, serTag, newpatches, stdout);
		}
		// ssearch <word [word1 [word2 ...]]>
		// Perform a keyword search integrated with related packages
		else if (opts.foundCommand() == opts.smartsearch)
		{
			auto_ptr<Ept> ept = debtagsInit();

			if (!opts.hasNext())
				throw wibble::exception::BadOption("you should specify one pattern to search");

			string keywords;
			while (opts.hasNext())
				if (keywords.empty())
					keywords = opts.next();
				else
					keywords += " " + opts.next();

			SmartSearcher searcher(*ept, keywords);

			if (opts.smse_reltags->boolValue())
				searcher.outputRelevantTags();
			else if (opts.smse_disctags->boolValue())
				searcher.outputDiscriminantTags();
			else {
				searcher.interact();
			}
		}
		// update
		// Updates the package tag database (requires root)
		else if (opts.foundCommand() == opts.update)
		{
			using namespace wibble::sys;

			verbose("System source directory: %s\n", ept::Path::debtagsSourceDir().c_str());
			verbose("User source directory: %s\n", ept::Path::debtagsUserSourceDir().c_str());

			if (!opts.misc_reindex->boolValue())
			{
				// Run the fetcher to acquire new data
				string fetcher = SCRIPTDIR "/fetch";
				if (!fs::access(fetcher, X_OK))
					warning("Fetch script %s does not exist or is not executable: skipping acquiring of new data\n", fetcher.c_str());
				else {
					if (opts.out_verbose->boolValue())
						fetcher += " --verbose";
					if (opts.misc_local->boolValue())
						fetcher += " --local";
					if (system(fetcher.c_str()) != 0)
						throw wibble::exception::Consistency("acquiring new data", "fetcher command " + fetcher + " failed");
				}
			}


			// Access the indexes to trigger a rebuild
			typedef ept::t::cache::debtags::IndexManager<> IndexManager;
			std::string a, b;

			auto_ptr<Ept> ept = debtagsInit();

			IndexManager::obtainWorkingPkgidx<ept::configuration::Apt>(ept->aggregator(), a);
			verbose("Package ID index: %s\n", a.c_str());

			IndexManager::obtainWorkingVocabulary(a, b);
			verbose("Vocabulary: %s\n", a.c_str());
			verbose("Vocabulary index: %s\n", b.c_str());

			IndexManager::obtainWorkingTagdb<ept::configuration::Apt>(ept->aggregator(), a, b);
			verbose("Tag database: %s\n", a.c_str());
			verbose("Tag database index: %s\n", b.c_str());

			//mode_t prev_umask = umask(022);
			//umask(prev_umask);
		}
		else if (opts.foundCommand() == opts.vocfilter)
		{
			auto_ptr<Ept> ept = debtagsInit();

			if (!opts.hasNext())
				throw wibble::exception::BadOption("you should specify the file with the collection to check");

			string file = opts.next();

			if (opts.misc_vocfile->boolValue())
			{
				ept::t::cache::debtags::VocabularyMerger vm;
				input::Stdio input(opts.misc_vocfile->stringValue());
				vm.read(input);
				ept->readCollectionRaw(file, vocabularyFilter(vm, textformat::OstreamWriter(cout)));
			}
			else
				ept->readCollectionRaw(file, vocabularyFilter(ept->vocabulary(), textformat::OstreamWriter(cout)));
		}
#if 0
		// todoreport
		// Print a report of packages needing work
		else if (opts.foundCommand() == opts.todoreport)
		{
			debtagsInit();
			wantTagDatabase();

			unsigned int itemsPerGroup = 0;
			if (opts.hasNext())
				itemsPerGroup = atoi(opts.next().c_str());

			ReportMaker rm(itemsPerGroup);
			rm.printReport();

		}
#endif
#if 0
		// install [-v] [-q] <tag expression>
		// apt-get install the packages that match the given tag expression
		else if (opts.foundCommand() == opts.install)
		{
			component::PackageTags& debtags = debtagsInit();
			wantTagDatabase();
			Installer installer;

			Searcher searcher(debtags, &installer);
			string expression = opts.next();

			searcher.output(expression, opts.matchGroup.invert->boolValue());

			return 1;
		}
#endif
		else
			throw wibble::exception::BadOption(string("unhandled command ") +
						(opts.foundCommand() ? opts.foundCommand()->name() : "(null)"));

		return 0;
	} catch (wibble::exception::BadOption& e) {
		cerr << e.desc() << endl;
		opts.outputHelp(cerr);
		return 1;
	} catch (std::exception& e) {
		cerr << e.what() << endl;
		return 1;
	}
}

#include <ept/cache/apt/packages.tcc>
#include <ept/cache/debtags/tagmap.tcc>
#include <tagcoll/coll/fast.tcc>
#include "Ept.tcc"

// vim:set ts=4 sw=4:
